#include "framecpp/config.h"

#include <iostream>

#include "framecpp/Common/CheckSumFilter.hh"
#include "framecpp/Common/IOStream.hh"
#include "framecpp/Common/Description.hh"
#include "framecpp/Common/Verify.hh"

#include "framecpp/Version4/FrEndOfFile.hh"
#include "framecpp/Version4/FrSE.hh"
#include "framecpp/Version4/FrSH.hh"

#include "framecpp/Version4/FrameSpec.hh"

using namespace FrameCPP::Version_4;

using FrameCPP::Common::CheckSum;
using FrameCPP::Common::CheckSumFilter;
using FrameCPP::Common::Description;
using FrameCPP::Common::FrameSpec;
using FrameCPP::Common::IStream;
using FrameCPP::Common::OStream;


FrEndOfFile::
FrEndOfFile( )
  : Common::FrEndOfFile( StructDescription( ) ),
    m_nFrames( 0 ),
    m_nBytes( 0 ),
    m_chkFlag( 0 ),
    m_chkSum( 0 ),
    m_seekTOC( 0 )
{
}

FrEndOfFile::
FrEndOfFile( IStream& Stream )
  : Common::FrEndOfFile( StructDescription( ) )
{
  //---------------------------------------------------------------------
  // :TODO: Need to handle calculation of checksums
  //---------------------------------------------------------------------
  Stream >> m_nFrames
	 >> m_nBytes
	 >> m_chkFlag
    ;
  //---------------------------------------------------------------------
  // Check for checksum calculations
  //---------------------------------------------------------------------
  CheckSumFilter*	crc = Stream.GetCheckSumFile( );

  if ( crc )
  {
   static const chkSum_type		chkSum = 0;

    crc->Filter( &chkSum, sizeof( chkSum ) );	// Use 0 value
    Stream.rdbuf( )->FilterRemove( crc );
  }
  Stream >> m_chkSum
    ;
  if ( crc )
  {
    Stream.rdbuf( )->FilterAdd( crc );
  }
  Stream >> m_seekTOC
    ;
  if ( crc )
  {
    crc->Value( );	// Stop processing
  }
}

FrEndOfFile::
~FrEndOfFile( )
{
}

FrameSpec::size_type FrEndOfFile::
Bytes( const Common::StreamBase& Stream ) const
{
  return
    sizeof( m_nFrames )
    + sizeof( m_nBytes )
    + sizeof( m_chkFlag )
    + sizeof( m_chkSum )
    + sizeof( m_seekTOC )
    ;
}

#if 0
void FrEndOfFile::
Read( IStream& Stream )
{
  //---------------------------------------------------------------------
  // :TODO: Need to calculate chkSum
  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  // Read data from stream
  //---------------------------------------------------------------------
  Stream >> m_data.nFrames
	 >> m_data.nBytes
	 >> m_data.chkFlag
	 >> m_data.chkSum
	 >> m_data.seekTOC
    ;
}
#endif /* 0 */

FrEndOfFile* FrEndOfFile::
Create( Common::IStream& Stream ) const
{
  return new FrEndOfFile( Stream );
}

const char* const FrEndOfFile::
ObjectStructName( ) const
{
  return StructName( );
}

const Description* FrEndOfFile::
StructDescription( )
{
  static Description ret;
  if ( ret.size( ) == 0 )
  {
    ret( FrSH( "FrEndOfFile", FrEndOfFile::s_object_id,
	       "End of File Data Structure" ) );
    ret( FrSE( "nFrames", "INT_4U",
	       "Number of frames in this file") );
    ret( FrSE( "nBytes", "INT_4U",
	       "Total number of bytes in this file"
	       " ( 0 if NOT computed )") );
    ret( FrSE( "chkFlag", "INT_4U",
	       "Checksum schemes: 0=none, 1=CRC, >2 presently unsued" ) );
    ret( FrSE ( "chkSum", "INT_4U", "File checksum." ) );
    ret( FrSE ( "seekTOC", "INT_4U",
		"Byes to back up to the beginning of the table of"
		" contents structure. If seekTOC == 0, then there"
		" is no TOC for this file." ) );
  }

  return &ret;
}

void FrEndOfFile::
VerifyObject( Common::Verify& Verifier,
	      Common::IFrameStream& Stream ) const
{
  Verifier.SeenFrEndOfFile( true );
  Verifier.ExamineFrEndOfFileChecksum( Stream, m_chkFlag, m_chkSum );
}
		  
void FrEndOfFile::
Write( OStream& Stream ) const
{
  //---------------------------------------------------------------------
  // Get local information
  //---------------------------------------------------------------------

  const nFrames_type	nFrames( Stream.GetNumberOfFrames( ) );
  const nBytes_type	nBytes( Stream.tellp( )
				+ std::streampos( Bytes( Stream ) ) );
  const seekTOC_type	seekTOC( nBytes - Stream.GetTOCOffset( ) );

  chkFlag_type		chkFlag = CheckSum::NONE;
  chkSum_type		chkSum = 0;

  //---------------------------------------------------------------------
  // Calculate checksum
  //---------------------------------------------------------------------
  CheckSumFilter*
    crc = Stream.GetCheckSumFile( );

  if ( crc )
  {
    chkFlag = crc->Type( );

    crc->Filter( &nFrames, sizeof( nFrames ) );
    crc->Filter( &nBytes, sizeof( nBytes ) );
    crc->Filter( &chkFlag, sizeof( chkFlag ) );
    crc->Filter( &chkSum, sizeof( chkSum ) );
    crc->Filter( &seekTOC, sizeof( seekTOC ) );
    
    chkSum = crc->Value( );
    
    Stream.SetCheckSumFile( CheckSum::NONE );
  }

  //---------------------------------------------------------------------
  // Write out to the stream
  //---------------------------------------------------------------------
  Stream << nFrames
	 << nBytes
	 << chkFlag
	 << chkSum
	 << seekTOC
    ;
}

#if 0
void FrEndOfFile::
Write( OFrameStreamGen& Stream ) const
{
#if 0
  CheckSumCRCFilter* crc
    = dynamic_cast< OFrameStream& >( Stream ).GetFrameCheckSumFilter( );

  if ( crc )
  {
    //-------------------------------------------------------------------
    // Fill a buffer with the appropriate data for calculating
    //   checksum.
    //-------------------------------------------------------------------
    static const size_t
      buf_size =
      sizeof( nFrames )
      + sizeof( nBytes )
      + sizeof( chkFlag )
      + sizeof( chkSum )
      + sizeof( seekTOC );

    char buffer[ buf_size ];
    char* pos = buffer;

    chkFlag = Common::CheckSum::CRC;
    chkSum = 0;

#define BUFFER_COPY( MEMBER ) memcpy( pos,&(MEMBER), sizeof( MEMBER ) ); \
    pos += sizeof( MEMBER )

    BUFFER_COPY( nFrames );
    BUFFER_COPY( nBytes );
    BUFFER_COPY( chkFlag );
    BUFFER_COPY( chkSum );
    BUFFER_COPY( seekTOC );

#undef BUFFER_COPY

    //-------------------------------------------------------------------
    // Calculate the checksum for this structure.
    //-------------------------------------------------------------------

    crc->pubfilter( buffer, buf_size );

    //-------------------------------------------------------------------
    // Set the chkSum value
    //-------------------------------------------------------------------
    chkSum = crc->Value( );
  }
  else
#endif /* 0 */
  {
    //-------------------------------------------------------------------
    // Fill in some information
    //-------------------------------------------------------------------
    chkFlag = Common::CheckSum::NONE;
    chkSum = 0;
  }

  // Write contents to the file
  Stream << nFrames
	 << nBytes
	 << chkFlag
	 << chkSum
	 << seekTOC;
}
#endif /* 0 */

bool FrEndOfFile::
operator==( const FrEndOfFile& RHS ) const
{
  return ( ( m_nFrames == RHS.m_nFrames )
	   && ( m_nBytes == RHS.m_nBytes )
	   && ( m_chkFlag == RHS.m_chkFlag )
	   && ( m_chkSum == RHS.m_chkSum )
	   && ( m_seekTOC == RHS.m_seekTOC ) );
}

bool FrEndOfFile::
operator==( const Common::FrameSpec::Object& Obj ) const
{
  return ( *this == Obj );
}

std::auto_ptr< FrameSpec::Object > FrEndOfFile::
demote( INT_2U Target, IStream* Stream ) const
{
  throw
    Unimplemented( "Object* FrEndOfFile::demote( Object* Obj ) const",
		   DATA_FORMAT_VERSION, __FILE__, __LINE__ );
}

std::auto_ptr< FrameSpec::Object > FrEndOfFile::
promote( INT_2U Target, FrameSpec::Object* Obj, IStream* Stream ) const
{
  throw
    Unimplemented( "Object* FrEndOfFile::promote( Object* Obj ) const",
		   DATA_FORMAT_VERSION, __FILE__, __LINE__ );
}
